""" A pymbolic node representing a sum factorization kernel """

from pytools import ImmutableRecord

import pymbolic.primitives as prim
import loopy as lp
import inspect


class SumfactKernel(ImmutableRecord, prim.Variable):
    def __init__(self,
                 matrix_sequence=None,
                 buffer=None,
                 stage=1,
                 preferred_position=None,
                 restriction=0,
                 within_inames=(),
                 input=None,
                 padding=frozenset(),
                 index=None,
                 insn_dep=frozenset(),
                 coeff_func=None,
                 element=None,
                 component=None,
                 accumvar=None,
                 ):
        """Create a sum factorization kernel

        Sum factorization can be written as

        Y = R_{d-1} (A_{d-1} * ... * R_0 (A_0 * X)...)

        with:
        - X: Input rank d tensor of dimension n_0 x ... x n_{d-1}
        - Y: Output rank d tensor of dimension m_0 x ... x m_{d-1}
        - A_l: Values of 1D basis evaluations at quadrature points in l
               direction, matrix of dimension m_l x n_l
        - R_l: Transformation operator that permutes the underlying data
               vector of the rank d tensor in such a way that the fastest
               direction gets the slowest direction

        In the l'th step we have the following setup:
        - A_l: Matrix of dimensions m_l x n_l
        - X_l: Rank d tensor of dimensions n_l x ... x n_{d-1} x m_0 x ... x m_{l-1}
        - R_l: Transformation operator

        Looking at the indizes the following will happen:
        X --> [n_l,...,n_{d-1},m_0,...,m_{l-1}]
        A_l * X --> [m_l,n_l] * [n_l, ...] = [m_l,n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]
        R_l (A_l*X) --> [n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]

        So the multiplication with A_l is a reduction over one index and
        the transformation brings the next reduction index in the fastest
        position.

        It can make sense to permute the order of directions. If you have
        a small m_l (e.g. stage 1 on faces) it is better to do direction l
        first. This can be done by:

        - Permuting the order of the A matrices.
        - Permuting the input tensor.
        - Permuting the output tensor (this assures that the directions of
          the output tensor are again ordered from 0 to d-1).

        Note, that you will typically *not* set all of the below arguments,
        but only some. The vectorization strategy may set others for you.
        The only argument really needed in all cases is matrix_sequence.

        Arguments:
        ----------
        matrix_sequence: A tuple of BasisTabulationMatrixBase instances
            The list of tensors to be applied to the input.
            Order of application is from 0 up.
        buffer: A string identifying the flip flop buffer in use
            for intermediate results. The memory is expected to be
            pre-initialized with the input or you have to provide
            direct_input (FastDGGridOperator).
        stage: 1 or 3
        insn_dep: An instruction ID that the first issued instruction
            should depend upon. All following ones will depend on each
            other.
        within_inames: Instructions will be executed within those
            inames (needed for stage 3 in jacobians).
        preferred_position: Will be used in the dry run to order kernels
            when doing vectorization e.g. (dx u,dy u,dz u, u).
        restriction: Restriction for faces values.
        input: The name to use for the input temporary
        padding: a set of slots in the input temporary to be padded
        index: The slot in the input temporary dedicated to this kernel
        coeff_func: The function to call to get the input coefficient
        element: The UFL element
        component: The treepath to the correct component of above element
        accumvar: The accumulation variable to accumulate into
        """
        # Check the input and apply defaults where necessary
        assert isinstance(matrix_sequence, tuple)
        from dune.perftool.sumfact.tabulation import BasisTabulationMatrixBase
        assert all(isinstance(m, BasisTabulationMatrixBase) for m in matrix_sequence)

        assert stage in (1, 3)

        if preferred_position is not None:
            assert isinstance(preferred_position, int)

        if stage == 3:
            assert isinstance(restriction, tuple)

        assert isinstance(within_inames, tuple)

        assert isinstance(insn_dep, frozenset)

        # The following construction is a bit weird: Dict comprehensions do not have
        # access to the locals of the calling scope: So we need to do the eval beforehand
        defaults = [eval(a) for a in SumfactKernel.init_arg_names]
        defaultdict = {a: defaults[SumfactKernel.init_arg_names.index(a)] for a in SumfactKernel.init_arg_names}

        # Call the base class constructors
        ImmutableRecord.__init__(self, **defaultdict)
        prim.Variable.__init__(self, "SUMFACT")

    #
    # The methods/fields needed to get a well-formed pymbolic node
    #

    def __getinitargs__(self):
        return tuple(getattr(self, arg) for arg in SumfactKernel.init_arg_names)

    def stringifier(self):
        return lp.symbolic.StringifyMapper

    mapper_method = "map_sumfact_kernel"

    #
    # Some cache key definitions
    # Watch out for the documentation to see which key is used unter what circumstances
    #

    @property
    def cache_key(self):
        """ The cache key that can be used in generation magic
        Any two sum factorization kernels having the same cache_key
        are realized simulatenously!
        """
        return (self.matrix_sequence, self.restriction, self.stage, self.buffer)

    @property
    def input_key(self):
        """ A cache key for the input coefficients
        Any two sum factorization kernels having the same input_key
        work on the same input coefficient (and are suitable for simultaneous
        treatment because of that)
        """
        return (self.restriction, self.stage, self.coeff_func, self.element, self.component, self.accumvar)

    #
    # Some convenience methods to extract information about the sum factorization kernel
    #

    @property
    def length(self):
        """ The number of matrices to apply """
        return len(self.matrix_sequence)

    @property
    def vectorized(self):
        return next(iter(self.matrix_sequence)).vectorized

    @property
    def transposed(self):
        return next(iter(self.matrix_sequence)).transpose

    @property
    def vec_index(self):
        """ A tuple with the vector index

        Defaults to the empty tuple to allow addition to any
        existing index tuple """
        if self.index is not None:
            return (self.index,)
        else:
            return ()

    @property
    def flat_input_shape(self):
        """ The 'flat' input tensor shape """
        from pytools import product
        shape = (product(mat.cols for mat in self.matrix_sequence),)
        if self.vectorized:
            shape = shape + (4,)
        return shape

    @property
    def quadrature_shape(self):
        """ The shape of a temporary for the quadrature points

        Takes into account the lower dimensionality of faces and vectorization.
        """
        if self.transposed:
            shape = tuple(mat.cols for mat in self.matrix_sequence if mat.face is None)
        else:
            shape = tuple(mat.rows for mat in self.matrix_sequence if mat.face is None)
        if self.vectorized:
            shape = shape + (4,)
        return shape

    @property
    def quadrature_dimtags(self):
        """ The dim_tags of a temporary for the quadrature points

        Takes into account the lower dimensionality of faces and vectorization.
        """
        tags = ["f"] * len(self.quadrature_shape)
        if self.vectorized:
            tags[-1] = 'c'
        return ",".join(tags)

    @property
    def dof_shape(self):
        """ The shape of a temporary for the degrees of freedom

        Takes into account vectorization.
        """
        shape = tuple(mat.rows for mat in self.matrix_sequence)
        if self.vectorized:
            shape = shape + (4,)
        return shape

    @property
    def dof_dimtags(self):
        """ The dim_tags of a temporary for the degrees of freedom

        Takes into account vectorization.
        """
        tags = ["f"] * len(self.dof_shape)
        if self.vectorized:
            tags[-1] = 'vec'
        return ",".join(tags)

    @property
    def output_shape(self):
        if self.stage == 1:
            return self.quadrature_shape
        else:
            return self.dof_shape

    @property
    def output_dimtags(self):
        if self.stage == 1:
            return self.quadrature_dimtags
        else:
            return self.dof_dimtags

    def output_to_pymbolic(self, name):
        if self.vectorized:
            return lp.TaggedVariable(name, "vector")
        else:
            return lp.TaggedVariable(name, "sumfac")


# Extract the argument list and store it on the class. This needs to be done
# outside of the class because the SumfactKernel class object needs to be fully
# initialized in order to extract the information from __init__.
SumfactKernel.init_arg_names = tuple(inspect.getargspec(SumfactKernel.__init__)[0][1:])
